@@ -404,12 +404,30 @@ class GitHandler(object):
 
             changed_refs = [ref for ref, sha in compat.iteritems(new_refs)
                             if sha != old_refs.get(ref)]
-            new = [bin(self.map_hg_get(new_refs[ref])) for ref in changed_refs]
+
+            new = [
+                bin(sha) for sha in map(self.map_hg_get, changed_refs) if sha
+            ]
             old = {}
-            for r in old_refs:
-                old_ref = self.map_hg_get(old_refs[r])
-                if old_ref:
-                    old[bin(old_ref)] = 1
+
+            for ref, sha in compat.iteritems(old_refs):
+                try:
+                    gittag = self.git.get_object(sha)
+                except KeyError:
+                    gittag = None
+
+                # make sure we don't accidentally dereference and lose
+                # annotated tags
+                if isinstance(gittag, Tag):
+                    sha = gittag.object[1]
+
+                old_target = self.map_hg_get(sha)
+
+                if old_target:
+                    self.ui.debug(b'unchanged ref %s: %s\n' % (ref, old_target))
+                    old[bin(old_target)] = 1
+                else:
+                    self.ui.debug(b'changed ref %s\n' % (ref))
 
             return old, new
         except (HangupException, GitProtocolError) as e:
@@ -1149,7 +1167,7 @@ class GitHandler(object):
         # the future.)
         if not refs or next(iter(refs.keys())) == b'capabilities^{}':
             if not exportable:
-                tip = self.repo.lookup(b'tip')
+                tip = self.repo.filtered(b'served').lookup(b'tip')
                 if tip != nullid:
                     if b'capabilities^{}' in new_refs:
                         del new_refs[b'capabilities^{}']
@@ -1192,7 +1210,11 @@ class GitHandler(object):
 
             for ref in rev_refs:
                 if ref not in refs:
-                    new_refs[ref] = self.map_git_get(ctx.hex())
+                    gitobj = self.git.get_object(self.git.refs[ref])
+                    if isinstance(gitobj, Tag):
+                        new_refs[ref] = gitobj.id
+                    else:
+                        new_refs[ref] = self.map_git_get(ctx.hex())
                 elif new_refs[ref] in self._map_git:
                     rctx = unfiltered[self.map_hg_get(new_refs[ref])]
                     if rctx.ancestor(ctx) == rctx or force:
@@ -1378,18 +1400,38 @@ class GitHandler(object):
             if self.repo.tagtype(tag) in (b'global', b'git'):
                 tag = tag.replace(b' ', b'_')
                 target = self.map_git_get(hex(sha))
-                if target is not None:
-                    tag_refname = b'refs/tags/' + tag
-                    if check_ref_format(tag_refname):
-                        self.git.refs[tag_refname] = target
-                        self.tags[tag] = hex(sha)
-                    else:
-                        self.repo.ui.warn(b"warning: not exporting tag '%s' "
-                                          b"due to invalid name\n" % tag)
-                else:
+
+                if target is None:
                     self.repo.ui.warn(b"warning: not exporting tag '%s' "
                                       b"due to missing git "
                                       b"revision\n" % tag)
+                    continue
+
+                tag_refname = b'refs/tags/' + tag
+
+                if not check_ref_format(tag_refname):
+                    self.repo.ui.warn(b"warning: not exporting tag '%s' "
+                                      b"due to invalid name\n" % tag)
+                    continue
+
+                # check whether the tag already exists and is
+                # annotated
+                if tag_refname in self.git.refs:
+                    gittarget = self.git.refs[tag_refname]
+                    gittag = self.git.get_object(gittarget)
+                    if isinstance(gittag, Tag):
+                        if gittag.object[1] != target:
+                            self.repo.ui.warn(
+                                b"warning: not overwriting annotated "
+                                b"tag '%s'\n" % tag
+                            )
+
+                        # never overwrite annotated tags, otherwise
+                        # it'd happen on every pull
+                        target = gittarget
+
+                self.git.refs[tag_refname] = target
+                self.tags[tag] = hex(sha)
 
     def _filter_for_bookmarks(self, bms):
         if not self.branch_bookmark_suffix:
@@ -1840,7 +1882,7 @@ class GitHandler(object):
 
             # not available in dulwich 0.19, used on Python 2.7
             if hasattr(client, 'get_credentials_from_store'):
-                urlobj = hgutil.url(uri)
+                urlobj = compat.url(uri)
                 auth = client.get_credentials_from_store(
                     urlobj.scheme,
                     urlobj.host,
@@ -1874,7 +1916,7 @@ class GitHandler(object):
             )
 
         if uri.startswith(b'file://'):
-            return client.LocalGitClient(), hgutil.url(uri).path
+            return client.LocalGitClient(), compat.url(uri).path
 
         # if its not git or git+ssh, try a local url..
         return client.SubprocessGitClient(), uri
